// libtype.cpp
// OpenType utilities
// Copyright (C)2000 Y.Yamakawa

#include "libtype.h"

const char *tags[]={
  "cmap","head","hhea","hmtx","maxp","name","OS/2","post", /* <- Required */
  "cvt ","fpgm","glyf","loca","prep",                      /* <- Related to TrueType Outlines */
  "CFF ","fvar","MMSD","MMFX",                             /* <- Related to Postscript Outlines */
  "BASE","GDEF","GPOS","GSUB","JSTF",                      /* <- Advanced Typographic */
  "EBDT","EBLC","EBSC",                                    /* <- Related to Bitmap Glyphs */
  "DSIG","gasp","hdmx","kern","LTSH","PCLT","VDMX","vhea","vmtx" /* Other OpenType Tables */
};

const char *behaviorID[]={"none","GASP_GRIDFIT","GASP_DOGRAY","GASP_GRIDFIT|GASP_DOGRAY"};
const char *platformID[]={"Apple Unicode","Macintosh","ISO","Microsoft"};

fsTypeID_type fsTypeID;
fsTypeOptionID_type fsTypeOptionID;
usWeightClassID_type usWeightClassID;
usWidthClassID_type usWidthClassID;
panFamilyID_type panFamilyID;
panSerifID_type panSerifID;
panWeightID_type panWeightID;
panProportionID_type panProportionID;
panContrastID_type panContrastID;
panStrokeID_type panStrokeID;
panArmID_type panArmID;
panLetterID_type panLetterID;
panMidlineID_type panMidlineID;
panXHeightID_type panXHeightID;

void initializeOTF(void){
  static bool flag=false;

  if(flag)
    return;
  flag=true;

  fsTypeID["CZX"]=0x0002;
  fsTypeID["\E\"]=0x0004;
  fsTypeID["ҏW\"]=0x0008;
  fsTypeID["\EEҏW\"]=0x000c;
  fsTypeID["CXg[\"]=0x0000;
  fsTypeOptionID["TuZbg"]=0x0100;
  fsTypeOptionID["rbg}bv̂"]=0x0200;

  usWeightClassID[100]="Thin";
  usWeightClassID[200]="Extra-light";
  usWeightClassID[300]="Light";
  usWeightClassID[400]="Regular";
  usWeightClassID[500]="Medium";
  usWeightClassID[600]="Semi-bold";
  usWeightClassID[700]="Bold";
  usWeightClassID[800]="Extra-bold";
  usWeightClassID[900]="Black";

  usWidthClassID[1]="Ultra-condensed";
  usWidthClassID[2]="Extra-condensed";
  usWidthClassID[3]="Condensed";
  usWidthClassID[4]="Semi-condensed";
  usWidthClassID[5]="Medium";
  usWidthClassID[6]="Semi-expanded";
  usWidthClassID[7]="Expanded";
  usWidthClassID[8]="Extra-expanded";
  usWidthClassID[9]="Ultra-expanded";

  panWeightID[0]="Any";
  panWeightID[1]="No Fit";
  panWeightID[2]="Very Light";
  panWeightID[3]="Light";
  panWeightID[4]="Thin";
  panWeightID[5]="Book";
  panWeightID[6]="Medium";
  panWeightID[7]="Demi";
  panWeightID[8]="Bold";
  panWeightID[9]="Heavy";
  panWeightID[10]="Black";
  panWeightID[11]="Nord";
}

OTF::OTF(const unsigned char *buf){
  int size=sizeof(sfnt_version)+sizeof(numTables)+sizeof(searchRange)+sizeof(entrySelector)+sizeof(rangeShift);
  int i,j;
  const unsigned char *buf2=buf;

  memcpy(&sfnt_version,buf,size);

  numTables=swap2(numTables);
  searchRange=swap2(searchRange);
  entrySelector=swap2(entrySelector);
  rangeShift=swap2(rangeShift);

  header=new THeader[numTables];

  correct=true;
  buf2+=size;
  for(i=0; i<numTables; i++){
    memcpy(header[i].tag,buf2,4);
    header[i].tag[4]='\0';
    buf2+=4;
    header[i].sum=swap4(*(((unsigned int *)buf2)++));
    header[i].offset=swap4(*(((unsigned int *)buf2)++));
    header[i].length=swap4(*(((unsigned int *)buf2)++));
/*
    if(tableSum(header+i)!=false){
      correct=false;
      break;
    }
*/
    for(j=0; j<numtags; j++){
      if(strcmp(tags[j],header[i].tag)==0){
        switch(j){
        case 6:  // OS/2
          tables[header[i].tag]=new OS2(buf+header[i].offset);
          break;
        case 26: // gasp
          tables[header[i].tag]=new gaspHeader(buf+header[i].offset);
          break;
        default:
          tables[header[i].tag]=NULL;
          break;
        }
        j=numtags+1;//m̃e[ułƂ
      }
    }
    if(j==numtags) //sȎނ̃e[uꍇ
      tables[header[i].tag]=NULL;
  }
  return;
}

OTF::~OTF(){
  delete header;
  map<string,fontObject *>::iterator ifO;
  for(ifO=tables.begin(); ifO!=tables.end(); ++ifO)
    if((*ifO).second!=NULL) delete (*ifO).second;
}

bool OTF::findTable(const char *tag,fontObject * &fO){
  map<string,fontObject *>::iterator ifO=tables.find(tag);
  if(ifO!=tables.end()){
    fO=(*ifO).second;
    return true;
  }else
    return false;
}

bool OTF::updateImage(unsigned char *buf){
  map<string,fontObject *>::iterator it;

  for(int i=0; i<numTables; i++){
    it=tables.find(header[i].tag);
    if(it!=tables.end() && (*it).second!=NULL)
      (*it).second->updateImage(buf+header[i].offset);
  }

  return true;
}

///////////////////////////////

gaspHeader::gaspHeader(const unsigned char *buf){
  int i;
  int size=sizeof(version)+sizeof(numRanges);
  gaspRange gr;

  memcpy(&version,buf,size);

  version=swap2(version);
  numRanges=swap2(numRanges);

  for(i=0,buf+=size; i<numRanges; i++,buf+=sizeof(gr)){
    memcpy(&gr,buf,sizeof(gr));
    ranges[swap2(gr.rangeMaxPPEM)]=swap2(gr.rangeGaspBehavior);
  }
  correct=true;
  return;
}

bool gaspHeader::updateImage(unsigned char *buf){
  map<unsigned short,unsigned short>::iterator ir;

  *(((unsigned short *)buf)++)=swap2(version);
  *(((unsigned short *)buf)++)=swap2(numRanges);

  for(ir=ranges.begin(); ir!=ranges.end(); ++ir){
    *(((unsigned short *)buf)++)=swap2((*ir).first);
    *(((unsigned short *)buf)++)=swap2((*ir).second);
  }
  return true;
}

///////////////////////////////

OS2::OS2(const unsigned char *buf){
  version=swap2(*(((unsigned short *)buf)++));
  if(version>OS2_VERSION){
    correct=false;
    return;
  }

  xAvgCharWidth=swap2(*(((short *)buf)++));
  usWeightClass=swap2(*(((unsigned short *)buf)++));
  usWidthClass=swap2(*(((unsigned short *)buf)++));
  fsType=swap2(*(((short *)buf)++));
  ySubscriptXSize=swap2(*(((short *)buf)++));
  ySubscriptYSize=swap2(*(((short *)buf)++));
  ySubscriptXOffset=swap2(*(((short *)buf)++));
  ySubscriptYOffset=swap2(*(((short *)buf)++));
  ySuperscriptXSize=swap2(*(((short *)buf)++));
  ySuperscriptYSize=swap2(*(((short *)buf)++));
  ySuperscriptXOffset=swap2(*(((short *)buf)++));
  ySuperscriptYOffset=swap2(*(((short *)buf)++));
  yStrikeoutSize=swap2(*(((short *)buf)++));
  yStrikeoutPosition=swap2(*(((short *)buf)++));
  sFamilyClass=swap2(*(((short *)buf)++));
  panose=*(((PANOSE *)buf)++);
  ulUnicodeRange1=swap4(*(((unsigned int *)buf)++));
  ulUnicodeRange2=swap4(*(((unsigned int *)buf)++));
  ulUnicodeRange3=swap4(*(((unsigned int *)buf)++));
  ulUnicodeRange4=swap4(*(((unsigned int *)buf)++));
  memcpy(achVendID,buf,4);
  buf+=4;
  fsSelection=swap2(*(((unsigned short *)buf)++));
  usFirstCharIndex=swap2(*(((unsigned short *)buf)++));
  usLastCharIndex=swap2(*(((unsigned short *)buf)++));
  sTypoAscender=swap2(*(((short *)buf)++));
  sTypoDescender=swap2(*(((short *)buf)++));
  sTypoLineGap=swap2(*(((short *)buf)++));
  usWinAscent=swap2(*(((unsigned short *)buf)++));
  usWinDescent=swap2(*(((unsigned short *)buf)++));
  ulCodePageRange1=swap4(*(((unsigned int *)buf)++));
  ulCodePageRange2=swap4(*(((unsigned int *)buf)++));
  if(version>1){
    sxHeight=swap2(*(((short *)buf)++));
    sCapHeight=swap2(*(((short *)buf)++));
    usDefaultChar=swap2(*(((unsigned short *)buf)++));
    usBreakChar=swap2(*(((unsigned short *)buf)++));
    usMaxContext=swap2(*((unsigned short *)buf));
  }

  correct=true;
}

bool OS2::updateImage(unsigned char *buf){
  if(correct==true){
    *(((unsigned short *)buf)++)=swap2(version);
    *(((short *)buf)++)=swap2(xAvgCharWidth);
    *(((unsigned short *)buf)++)=swap2(usWeightClass);
    *(((unsigned short *)buf)++)=swap2(usWidthClass);
    *(((short *)buf)++)=swap2(fsType);
    *(((short *)buf)++)=swap2(ySubscriptXSize);
    *(((short *)buf)++)=swap2(ySubscriptYSize);
    *(((short *)buf)++)=swap2(ySubscriptXOffset);
    *(((short *)buf)++)=swap2(ySubscriptYOffset);
    *(((short *)buf)++)=swap2(ySuperscriptXSize);
    *(((short *)buf)++)=swap2(ySuperscriptYSize);
    *(((short *)buf)++)=swap2(ySuperscriptXOffset);
    *(((short *)buf)++)=swap2(ySuperscriptYOffset);
    *(((short *)buf)++)=swap2(yStrikeoutSize);
    *(((short *)buf)++)=swap2(yStrikeoutPosition);
    *(((short *)buf)++)=swap2(sFamilyClass);
    *(((PANOSE *)buf)++)=panose;
    *(((unsigned int *)buf)++)=swap4(ulUnicodeRange1);
    *(((unsigned int *)buf)++)=swap4(ulUnicodeRange2);
    *(((unsigned int *)buf)++)=swap4(ulUnicodeRange3);
    *(((unsigned int *)buf)++)=swap4(ulUnicodeRange4);
    memcpy(buf,achVendID,4);
    buf+=4;
    *(((unsigned short *)buf)++)=swap2(fsSelection);
    *(((unsigned short *)buf)++)=swap2(usFirstCharIndex);
    *(((unsigned short *)buf)++)=swap2(usLastCharIndex);
    *(((short *)buf)++)=swap2(sTypoAscender);
    *(((short *)buf)++)=swap2(sTypoDescender);
    *(((short *)buf)++)=swap2(sTypoLineGap);
    *(((unsigned short *)buf)++)=swap2(usWinAscent);
    *(((unsigned short *)buf)++)=swap2(usWinDescent);
    *(((unsigned int *)buf)++)=swap4(ulCodePageRange1);
    *(((unsigned int *)buf)++)=swap4(ulCodePageRange2);
    if(version>1){
      *(((short *)buf)++)=swap2(sxHeight);
      *(((short *)buf)++)=swap2(sCapHeight);
      *(((unsigned short *)buf)++)=swap2(usDefaultChar);
      *(((unsigned short *)buf)++)=swap2(usBreakChar);
      *((unsigned short *)buf)=swap2(usMaxContext);
    }
    return true;
  }else
    return false;
}
