// libtype.h
// OpenType utilities
// Copyright (C)2000 Y.Yamakawa

#ifndef _LIBTYPE_H
#define _LIBTYPE_H

#include <iostream.h>
#include <mem.h>

#ifdef __BCPLUSPLUS__
#include <string.h>   // stringNX
#include <algorith.h> // STLʃASY
#else
#include <bstring.h>
#include <algo.h>
#endif /* __BCPLUSPLUS__ */

#include <map.h>
#include "util.h"

// 萔FꕔinitializeOTF֐ŏB

extern const char *tags[];
#define numtags 35

typedef map<char *,unsigned short> fsTypeID_type;
extern fsTypeID_type fsTypeID;

typedef map<char *,unsigned short> fsTypeOptionID_type;
extern fsTypeOptionID_type fsTypeOptionID;

typedef map<unsigned short,char *> usWeightClassID_type;
extern usWeightClassID_type usWeightClassID;

typedef map<unsigned short,char *> usWidthClassID_type;
extern usWidthClassID_type usWidthClassID;

typedef map<unsigned char,char *> panFamilyID_type;
extern panFamilyID_type panFamilyID;
typedef map<unsigned char,char *> panSerifID_type;
extern panSerifID_type panSerifID;
typedef map<unsigned char,char *> panWeightID_type;
extern panWeightID_type panWeightID;
typedef map<unsigned char,char *> panProportionID_type;
extern panProportionID_type panProportionID;
typedef map<unsigned char,char *> panContrastID_type;
extern panContrastID_type panContrastID;
typedef map<unsigned char,char *> panStrokeID_type;
extern panStrokeID_type panStrokeID;
typedef map<unsigned char,char *> panArmID_type;
extern panArmID_type panArmID;
typedef map<unsigned char,char *> panLetterID_type;
extern panLetterID_type panLetterID;
typedef map<unsigned char,char *> panMidlineID_type;
extern panMidlineID_type panMidlineID;
typedef map<unsigned char,char *> panXHeightID_type;
extern panXHeightID_type panXHeightID;

// ֐FOTFNXgvO͍ŏ1x͌ĂяoƁB

void initializeOTF(void);


// NXE\̒`

struct THeader;

class fontObject{
protected:
  bool correct;
public:
  virtual ~fontObject(){};
  bool isCorrect();
  virtual bool updateImage(unsigned char *buf)=0;
  //virtual unsigned char *generateImage(void)=0;
  //virtual unsigned int getImageSize(void)=0;
};

class OTF : public fontObject{
  unsigned int sfnt_version;
  unsigned short numTables;
  unsigned short searchRange;
  unsigned short entrySelector;
  unsigned short rangeShift;
  THeader *header;
  map<string,fontObject *> tables;
public:
  OTF(const unsigned char *buf);
  ~OTF();
  bool findTable(const char *tag,fontObject * &fO);
  virtual bool updateImage(unsigned char *buf);
};

struct THeader{
  char tag[5];
  unsigned int sum;
  unsigned int offset;
  unsigned int length;
};


// gasp֘A

#define GASP_DOGRAY 2
#define GASP_GRIDFIT 1
#define GASP_INFINITY 0xffff

extern const char *behaviorID[];

//struct gaspRangeBehavior;

struct gaspHeader : public fontObject{
  unsigned short version;
  unsigned short numRanges;  // ݒ͈͂̐
  map<unsigned short,unsigned short> ranges;  // ݒ͈
  gaspHeader(const unsigned char *buf);
  virtual bool updateImage(unsigned char *buf);
};

/*
struct gaspRangeBehavior{
  bool gridfit;  // trueȂqeBOL
  bool dogray;  // trueȂX[WOL
}
*/

struct gaspRange{  // 荞ݎɎgp^
  unsigned short rangeMaxPPEM;  // ͈͂̏
  unsigned short rangeGaspBehavior;  // ݒl
};

// name֘A

extern const char *platformID[];

struct nameHeader{
  unsigned short formatSelector;
  unsigned short numRecords;
  unsigned short offset;
};

struct nameRecord{
  unsigned short platformID;
  unsigned short encordingID;
  unsigned short languageID;
  unsigned short nameID;
  unsigned short length;
  unsigned short offset;
  string str;
};

// OS/2֘A

#define OS2_VERSION 0x0002

#ifndef WIN32
struct PANOSE{
  unsigned char bFamilyType;
  unsigned char bSerifStyle;
  unsigned char bWeight;
  unsigned char bProportion;
  unsigned char bContrast;
  unsigned char bStrokeVariation;
  unsigned char bArmStyle;
  unsigned char bLetterform;
  unsigned char bMidline;
  unsigned char bXHeight;
};
#endif

struct OS2 : public fontObject{
  unsigned short version; // == 0x0002
  short xAvgCharWidth;
  unsigned short usWeightClass;
  unsigned short usWidthClass;
  short fsType;
  short ySubscriptXSize;
  short ySubscriptYSize;
  short ySubscriptXOffset;
  short ySubscriptYOffset;
  short ySuperscriptXSize;
  short ySuperscriptYSize;
  short ySuperscriptXOffset;
  short ySuperscriptYOffset;
  short yStrikeoutSize;
  short yStrikeoutPosition;
  short sFamilyClass;
  PANOSE panose;
  unsigned int ulUnicodeRange1; // bit 0-31
  unsigned int ulUnicodeRange2; // bit 32-63
  unsigned int ulUnicodeRange3; // bit 64-95
  unsigned int ulUnicodeRange4; // bit 96-127
  char achVendID[4];
  unsigned short fsSelection;
  unsigned short usFirstCharIndex;
  unsigned short usLastCharIndex;
  short sTypoAscender;
  short sTypoDescender;
  short sTypoLineGap;
  unsigned short usWinAscent;
  unsigned short usWinDescent;
  unsigned int ulCodePageRange1; // bit 0-31
  unsigned int ulCodePageRange2; // bit 32-63
  short sxHeight;
  short sCapHeight;
  unsigned short usDefaultChar;
  unsigned short usBreakChar;
  unsigned short usMaxContext;

  OS2(const unsigned char *buf);
  virtual bool updateImage(unsigned char *buf);
};

#endif